from typing import TypedDict


class EventType(str):
    pass


class Events:
    class Connected(TypedDict):
        peer_id: str

    class Disconnected(TypedDict):
        peer_id: str

    class Speaking(TypedDict):
        peer_id: str

    class Listening(TypedDict):
        peer_id: str

    class Thinking(TypedDict):
        peer_id: str



import asyncio
import fractions
import logging
import threading

import numpy as np
from aiortc.mediastreams import MediaStreamError, MediaStreamTrack
from av import AudioFrame
from av.audio.fifo import AudioFifo
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class AudioTrackOptions(BaseModel):
    """Audio Track Options"""

    sample_rate: int = 24000
    """
    Sample Rate is the number of samples of audio carried per second, measured in Hz, Default is 24000.
    """

    channels: int = 1
    """
    Channels is the number of audio channels, Default is 1, which is mono.
    """

    sample_width: int = 2
    """
    Sample Width is the number of bytes per sample, Default is 2, which is 16 bits.
    """


class AudioTrack(MediaStreamTrack):
    kind = "audio"

    def __init__(self, options=AudioTrackOptions()):
        print("AudioTrack __init__")
        super().__init__()

        # Audio configuration
        self.sample_rate = options.sample_rate
        self.channels = options.channels
        self.sample_width = options.sample_width  # 2 bytes per sample (16 bits)

        self._start = None
        self._timestamp = 0

        self.AUDIO_PTIME = 0.020  # 20ms audio packetization
        self.frame_samples = int(self.AUDIO_PTIME * self.sample_rate)

        # Audio FIFO buffer
        self.audio_fifo = AudioFifo()
        self.fifo_lock = threading.Lock()

    def __repr__(self) -> str:
        return f"<AudioTrack kind={self.kind} state={self.readyState}> sample_rate={self.sample_rate} channels={self.channels} sample_width={self.sample_width}>"

    def enqueue_audio(self, audio_bytes: bytes):
        """Process and add audio data directly to the AudioFifo"""
        if self.readyState != "live":
            return

        try:
            audio_array = np.frombuffer(audio_bytes, dtype=np.int16)
            audio_array = audio_array.reshape(self.channels, -1)

            frame = AudioFrame.from_ndarray(
                audio_array,
                format="s16",
                layout="mono" if self.channels == 1 else "stereo",
            )
            frame.sample_rate = self.sample_rate
            frame.time_base = fractions.Fraction(1, self.sample_rate)

            with self.fifo_lock:
                self.audio_fifo.write(frame)

        except Exception as e:
            logger.error(f"Error in enqueue_audio: {e}", exc_info=True)

    def flush_audio(self):
        """Flush the audio FIFO buffer"""
        with self.fifo_lock:
            self.audio_fifo = AudioFifo()

    async def recv(self) -> AudioFrame:
        """Receive the next audio frame"""
        if self.readyState != "live":
            raise MediaStreamError

        if self._start is None:
            self._start = asyncio.get_event_loop().time()
            self._timestamp = 0

        samples = self.frame_samples
        self._timestamp += samples

        target_time = self._start + (self._timestamp / self.sample_rate)
        current_time = asyncio.get_event_loop().time()
        wait = target_time - current_time

        if wait > 0:
            await asyncio.sleep(wait)

        try:
            # Read frames from the FIFO buffer
            with self.fifo_lock:
                frame = self.audio_fifo.read(samples)

            if frame is None:
                # If no data is available, generate silence
                frame = AudioFrame(
                    format="s16",
                    layout="mono" if self.channels == 1 else "stereo",
                    samples=samples,
                )
                for p in frame.planes:
                    p.update(np.zeros(samples, dtype=np.int16).tobytes())

                frame.sample_rate = self.sample_rate
                frame.time_base = fractions.Fraction(1, self.sample_rate)
            else:
                # Update frame properties
                frame.sample_rate = self.sample_rate
                frame.time_base = fractions.Fraction(1, self.sample_rate)

            # Set frame PTS
            frame.pts = self._timestamp

            return frame

        except Exception as e:
            logger.error(f"Error in recv: {e}", exc_info=True)
            raise MediaStreamError("Error processing audio frame")

    def stop(self) -> None:
        """Stop the track"""
        if self.readyState == "live":
            super().stop()



from .agent import Agent, AgentOptions
from .rtc import RTC, RTCOptions

__all__ = [
    "Agent",
    "AgentOptions",
    "RTC",
    "RTCOptions",
]

# Cleanup docs of unexported modules
_module = dir()
NOT_IN_ALL = [m for m in _module if m not in __all__]

__pdoc__ = {}

for n in NOT_IN_ALL:
    __pdoc__[n] = False



import logging
from typing import Optional

from pydantic import BaseModel

from ..rtc import RTC, RTCOptions
from ..rtc.audio_track import AudioTrack
from ..utils.emitter import EnhancedEventEmitter
from ._exceptions import RoomNotConnectedError, RoomNotCreatedError


class AgentOptions(BaseModel):
    """ "
    Every Agent is created with a set of options that define the configuration for the Agent.

    Args:
        rtc_options (RTCOptions): RTC Options for the Agent.
        audio_track (MediaStreamTrack): Audio Stream Track for the Agent.
    """

    rtc_options: RTCOptions
    """
    RTC Options is the configuration for the RTC.
    """

    audio_track: Optional[AudioTrack]
    """
    Audio Track is the Audio Stream Track for the Agent.
    """

    class Config:
        arbitrary_types_allowed = True


logger = logging.getLogger("Agent")


class Agent(EnhancedEventEmitter):
    """
    Agents is defined as the higher level user which is its own entity and has exposed APIs to
    interact with different Models and Outer World using dRTC.

    Agents can be connected to the dRTC Network by two stepsL
        - Joining the Agent to the dRTC Network, which assigns the Agent to a Room, using this Agent can setup event Listeners before connecting to the Room.
        - Connecting the Agent to the Room, which allows the Agent to send and receive data from the Room, and become part of the Room.
    """

    def __init__(self, options: AgentOptions):
        super(Agent, self).__init__()

        # Options is the configuration for the Agent.
        self.options = options

        # RTC is the Real Time Communication Handler for the Agent.
        self.__rtc = RTC(options.rtc_options)

        # Audio Track is the Audio Stream Track for the Agent.
        self.audio_track = options.audio_track

        # Logger for the Agent.
        self._logger = logger.getChild("Agent")

    @property
    def rtc(self):
        return self.__rtc

    @property
    def logger(self):
        return self._logger

    @property
    def room(self):
        """
        Room is the the abstraction under which every participant and the Agent is connected to, This is only available after the Agent is connected.

        raise ValueError if Agent is not connected to the Room.
        """
        if not self.__rtc.room:
            raise RoomNotCreatedError()

        return self.__rtc.room

    async def join(self):
        """
        Joins the dRTC Network and creates a Room instance for the Agent.

        This method establishes a connection to the dRTC Network and creates a Room instance for the agent.
        The returned Room object allows you to listen to events such as:
        - Remote peers joining or leaving the Room.
        - Room-specific events triggered during its lifecycle.

        Example Usage:
            ```python
            room = await agent.join()


            @room.on(RoomEvents.RoomJoined)
            def on_room_join():
                print("Room successfully joined!")
            ```
        """
        self.logger.info("Joining Agent to the dRTC Network")

        room = await self.__rtc.join()

        if not room:
            raise RoomNotConnectedError()

        return room

    async def connect(self):
        """
        Connects the Agent to the Room, This is only available after the Agent is joined to the dRTC Network.
        """
        self.logger.info("Connecting Agent to the Room")

        room = self.__rtc.room

        if not room:
            raise RoomNotCreatedError()

        await room.connect()



from typing import Literal

from av import AudioFrame
from av import AudioResampler as Resampler
from av.audio.fifo import AudioFifo


class AudioResampler:
    """
    Audio Resampler is used to resample the audio to the desired format, and store it in the Audio FIFO Buffer.
    Using the `resample` method, you can resample the audio frame to the desired format, and using the `recv` method, you can get the resampled audio frame.
    """

    def __init__(
        self, format: Literal["s16"], layout: Literal["mono", "stereo"], rate: int
    ):
        self.audio_fifo = AudioFifo()
        """
        Audio FIFO Buffer for the Audio Resampler.
        """

        self.resampler = Resampler(
            format=format,
            layout=layout,
            rate=rate,
        )
        """
        For the Audio Resampling, which is used to resample the audio to the desired format.
        """

    def resample(self, audio_frame: AudioFrame):
        """
        Resample the audio frame to the desired format, and add it to the Audio FIFO
        You can use the `recv` method to get the resampled audio frame.
        """
        resampled_frames = self.resampler.resample(audio_frame)

        for frame in resampled_frames:
            self.audio_fifo.write(frame)

    def recv(self) -> None | bytes:
        """
        Receive the audio frame from the Audio Resampler, which are stored in the Audio FIFO Queue.
        """
        frame = self.audio_fifo.read()

        if frame is None:
            return None

        pcm_data = frame.to_ndarray()

        pcm_bytes = pcm_data.tobytes()

        return pcm_bytes

    def clear(self):
        """
        Clear the Audio FIFO Buffer.
        """
        self.audio_fifo.clear()



import asyncio
import logging
from typing import Dict

from aiortc.mediastreams import MediaStreamTrack

from ....rtc.audio_resampler import AudioResampler
from . import _exceptions

logger = logging.getLogger(__name__)


class Conversation:
    def __init__(self, id: str):
        self.id = id
        """
        Conversation ID for the Realtime Conversation.
        """

        self.audio_resampler = AudioResampler(format="s16", layout="mono", rate=16000)
        """
        Audio Resampler for the Realtime Conversation, takes in Audio Frame and resamples it to the desired format.
        """

        self._logger = logger.getChild("Conversation")
        """
        Logger for the Conversation.
        """

        self._track_fut: Dict[str, asyncio.Future] = {}
        """
        Track Futures for different tracks.
        """

        self._active = True
        """
        Is the Conversation Active.
        """

    def __str__(self):
        return f"Conversation ID: {self.id}"

    def __repr__(self):
        return f"Conversation ID: {self.id}"

    @property
    def logger(self):
        return self._logger

    @property
    def active(self):
        return self._active

    def add_track(self, id: str, track: MediaStreamTrack):
        """
        Add a Track to the Conversation, which streamlines conversation into one Audio Stream.
        which can be later retrieved using the `recv_frame` method and feeded to the Model.
        """
        if track.kind != "audio":
            raise _exceptions.RealtimeModelTrackInvalidError()

        if self._track_fut.get(id):
            raise _exceptions.RealtimeModelError("Track is already started.")

        async def handle_audio_frame():
            try:
                while self._active and track.readyState != "ended":
                    frame = await track.recv()

                    frame.pts = None

                    if frame is None:
                        continue

                    self.audio_resampler.resample(frame)
            except Exception as e:
                self.logger.error(f"Error in handling audio frame: {e}")

        self._started_fut = asyncio.create_task(
            handle_audio_frame(), name=f"Conversation-{id}"
        )

        self._track_fut[id] = self._started_fut

    def stop(self):
        """
        Stop the Conversation and clear the Audio FIFO Buffer.
        """
        self._active = False

        self.audio_resampler.clear()

    def recv(self):
        """
        Receive the resampled audio frame from the Audio Resampler.
        """
        return self.audio_resampler.recv()



from __future__ import annotations


class RealtimeModelError(Exception):
    """Base class for exceptions in this module."""

    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class RealtimeModelNotConnectedError(RealtimeModelError):
    """Exception raised for errors in the Realtime Model, when Model is not connected."""

    def __init__(
        self,
    ):
        super().__init__(
            """
            Model is not connected, Possible ways to fix this:
            - Call the `model.connect()` before calling any other method on the Model.
            - Make sure the API_KEY and Project_ID are valid before connecting to the Model.
            """
        )


class RealtimeModelSocketError(RealtimeModelError):
    """Exception raised for errors in the Realtime Model, when Socket Connection Errored"""

    def __init__(
        self,
    ):
        super().__init__(
            """
            Socket Connection Error, Possible ways to fix this:
            - Check the network connection and try again.
            - Make sure the RealTime API is up and running.
            """
        )


class RealtimeModelTrackInvalidError(RealtimeModelError):
    """Exception raised for errors in the Realtime Model, when Track is invalid"""

    def __init__(
        self,
    ):
        super().__init__(
            """
            Track is invalid, Possible ways to fix this:
            - Make sure the Track is of type `audio`.
            - Make sure the Track is not None.
            """
        )



from __future__ import annotations

from typing import Literal, Union

from typing_extensions import NotRequired, TypedDict

SAMPLE_RATE = 24000
NUM_CHANNELS = 1


class FunctionToolChoice(TypedDict):
    type: Literal["function"]
    name: str


# OpenAI Voices
Voice = Literal["alloy", "echo", "shimmer", "ash", "ballad", "coral", "sage", "verse"]
"""
Voices available for OpenAI API, Voices are used to generate audio.
"""

ToolChoice = Union[Literal["auto", "none", "required"], FunctionToolChoice]
"""
ToolChoice is used to specify the tool to use for the generation, e.g. Weather etc.
"""

Role = Literal["system", "assistant", "user", "tool"]
"""
Role is used to specify the role of the speaker in the conversation.
"""

ResponseFinishedReason = Literal["stop", "max_tokens", "content_filter", "interrupt"]
"""
ResponseFinishedReason is used to specify the reason for the response generation to stop.
"""

AudioFormat = Literal["pcm16"]
"""
AudioFormat is used to specify the audio format.
"""

TranscriptionModels = Literal["whisper-1"]
"""
TranscriptionModels is used to specify the model to use for the generation.
"""

RealTimeModels = Literal[
    "gpt-4o-realtime-preview", "gpt-4o-realtime", "gpt-4o-realtime-preview-2024-12-17"
]
"""
RealTimeModels is used to specify the model to use for the generation.
"""

Modality = Literal["text", "audio"]
"""
Modality is used to specify the modality of the input.
"""

ResponseStatus = Literal[
    "in_progress", "completed", "incomplete", "cancelled", "failed"
]
"""
ResponseStatus is used to specify the status of the response.
"""


class TextContent(TypedDict):
    type: Literal["text"]
    text: str


class InputTextContent(TypedDict):
    type: Literal["input_text"]
    text: str


class AudioContent(TypedDict):
    type: Literal["audio"]
    audio: str  # base64


class InputAudioContent(TypedDict):
    type: Literal["input_audio"]
    audio: str  # base64


Content = Union[InputTextContent, TextContent, AudioContent, InputAudioContent]


class ContentPart(TypedDict):
    type: Literal["text", "audio"]
    audio: NotRequired[str]  # b64
    transcript: NotRequired[str]


class InputAudioTranscription(TypedDict):
    model: TranscriptionModels | str


class InputRealTimeModel(TypedDict):
    model: RealTimeModels


class ServerVad(TypedDict):
    type: Literal["server_vad"]
    threshold: NotRequired[float]
    prefix_padding_ms: NotRequired[int]
    silence_duration_ms: NotRequired[int]


class FunctionTool(TypedDict):
    type: Literal["function"]
    name: str
    description: NotRequired[str | None]
    parameters: dict


class SystemItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["message"]
    role: Literal["system"]
    content: list[InputTextContent]


class UserItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["message"]
    role: Literal["user"]
    content: list[InputTextContent | InputAudioContent]


class AssistantItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["message"]
    role: Literal["assistant"]
    content: list[TextContent | AudioContent]


class FunctionCallItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["function_call"]
    call_id: str
    name: str
    arguments: str


class FunctionCallOutputItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["function_call_output"]
    call_id: str
    output: str


class CancelledStatusDetails(TypedDict):
    type: Literal["cancelled"]
    reason: Literal["turn_detected", "client_cancelled"]


class IncompleteStatusDetails(TypedDict):
    type: Literal["incomplete"]
    reason: Literal["max_output_tokens", "content_filter"]


class Error(TypedDict):
    code: str
    message: str


class FailedStatusDetails(TypedDict):
    type: Literal["failed"]
    error: NotRequired[Error | None]


ResponseStatusDetails = Union[
    CancelledStatusDetails, IncompleteStatusDetails, FailedStatusDetails
]


class InputTokenDetails(TypedDict):
    cached_tokens: int
    text_tokens: int
    audio_tokens: int


class OutputTokenDetails(TypedDict):
    text_tokens: int
    audio_tokens: int


class Usage(TypedDict):
    total_tokens: int
    input_tokens: int
    output_tokens: int
    input_token_details: InputTokenDetails
    output_token_details: OutputTokenDetails


class Resource:
    class Session(TypedDict):
        id: str
        object: Literal["realtime.session"]
        expires_at: int
        model: str
        modalities: list[Literal["text", "audio"]]
        instructions: str
        voice: Voice
        input_audio_format: AudioFormat
        output_audio_format: AudioFormat
        input_audio_transcription: InputAudioTranscription | None
        turn_detection: ServerVad | None
        tools: list[FunctionTool]
        tool_choice: ToolChoice
        temperature: float
        max_response_output_tokens: int | Literal["inf"]

    class Conversation(TypedDict):
        id: str
        object: Literal["realtime.conversation"]

    Item = Union[SystemItem, UserItem, FunctionCallItem, FunctionCallOutputItem]

    class Response(TypedDict):
        id: str
        object: Literal["realtime.response"]
        status: ResponseStatus
        status_details: NotRequired[ResponseStatusDetails | None]
        output: list[Resource.Item]
        usage: NotRequired[Usage | None]


class ClientEvent:
    class SessionUpdateData(TypedDict):
        modalities: list[Literal["text", "audio"]]
        instructions: str
        voice: Voice
        input_audio_format: AudioFormat
        output_audio_format: AudioFormat
        input_audio_transcription: InputAudioTranscription | None
        turn_detection: ServerVad | None
        tools: list[FunctionTool]
        tool_choice: ToolChoice
        temperature: float
        # microsoft does not support inf, but accepts None
        max_response_output_tokens: int | Literal["inf"] | None

    class SessionUpdate(TypedDict):
        event_id: NotRequired[str]
        type: Literal["session.update"]
        session: ClientEvent.SessionUpdateData

    class InputAudioBufferAppend(TypedDict):
        event_id: NotRequired[str]
        type: Literal["input_audio_buffer.append"]
        audio: str  # b64

    class InputAudioBufferCommit(TypedDict):
        event_id: NotRequired[str]
        type: Literal["input_audio_buffer.commit"]

    class InputAudioBufferClear(TypedDict):
        event_id: NotRequired[str]
        type: Literal["input_audio_buffer.clear"]

    class UserItemCreate(TypedDict):
        id: str | None
        type: Literal["message"]
        role: Literal["user"]
        content: list[InputTextContent | InputAudioContent]

    class AssistantItemCreate(TypedDict):
        id: str | None
        type: Literal["message"]
        role: Literal["assistant"]
        content: list[TextContent]

    class SystemItemCreate(TypedDict):
        id: str | None
        type: Literal["message"]
        role: Literal["system"]
        content: list[InputTextContent]

    class FunctionCallOutputItemCreate(TypedDict):
        id: str | None
        type: Literal["function_call_output"]
        call_id: str
        output: str

    class FunctionCallItemCreate(TypedDict):
        id: str | None
        type: Literal["function_call"]
        call_id: str
        name: str
        arguments: str

    ConversationItemCreateContent = Union[
        UserItemCreate,
        AssistantItemCreate,
        SystemItemCreate,
        FunctionCallOutputItemCreate,
        FunctionCallItemCreate,
    ]

    class ConversationItemCreate(TypedDict):
        event_id: NotRequired[str]
        type: Literal["conversation.item.create"]
        previous_item_id: NotRequired[str | None]
        item: ClientEvent.ConversationItemCreateContent

    class ConversationItemTruncate(TypedDict):
        event_id: NotRequired[str]
        type: Literal["conversation.item.truncate"]
        item_id: str
        content_index: int
        audio_end_ms: int

    class ConversationItemDelete(TypedDict):
        event_id: NotRequired[str]
        type: Literal["conversation.item.delete"]
        item_id: str

    class ResponseCreateData(TypedDict, total=False):
        modalities: list[Literal["text", "audio"]]
        instructions: str
        voice: Voice
        output_audio_format: AudioFormat
        tools: list[FunctionTool]
        tool_choice: ToolChoice
        temperature: float
        max_output_tokens: int | Literal["inf"]

    class ResponseCreate(TypedDict):
        event_id: NotRequired[str]
        type: Literal["response.create"]
        response: NotRequired[ClientEvent.ResponseCreateData]

    class ResponseCancel(TypedDict):
        event_id: NotRequired[str]
        type: Literal["response.cancel"]


class ServerEvent:
    class ErrorContent(TypedDict):
        type: str
        code: NotRequired[str]
        message: str
        param: NotRequired[str]
        event_id: NotRequired[str]

    class Error(TypedDict):
        event_id: str
        type: Literal["error"]
        error: ServerEvent.ErrorContent

    class SessionCreated(TypedDict):
        event_id: str
        type: Literal["session.created"]
        session: Resource.Session

    class SessionUpdated(TypedDict):
        event_id: str
        type: Literal["session.updated"]
        session: Resource.Session

    class ConversationCreated(TypedDict):
        event_id: str
        type: Literal["conversation.created"]
        conversation: Resource.Conversation

    class InputAudioBufferCommitted(TypedDict):
        event_id: str
        type: Literal["input_audio_buffer.committed"]
        item_id: str

    class InputAudioBufferCleared(TypedDict):
        event_id: str
        type: Literal["input_audio_buffer.cleared"]

    class InputAudioBufferSpeechStarted(TypedDict):
        event_id: str
        type: Literal["input_audio_buffer.speech_started"]
        item_id: str
        audio_start_ms: int

    class InputAudioBufferSpeechStopped(TypedDict):
        event_id: str
        type: Literal["input_audio_buffer.speech_stopped"]
        item_id: str
        audio_end_ms: int

    class ConversationItemCreated(TypedDict):
        event_id: str
        type: Literal["conversation.item.created"]
        previous_item_id: str | None
        item: Resource.Item

    class ConversationItemInputAudioTranscriptionCompleted(TypedDict):
        event_id: str
        type: Literal["conversation.item.input_audio_transcription.completed"]
        item_id: str
        content_index: int
        transcript: str

    class InputAudioTranscriptionError(TypedDict):
        type: str
        code: NotRequired[str]
        message: str
        param: NotRequired[str]

    class ConversationItemInputAudioTranscriptionFailed(TypedDict):
        event_id: str
        type: Literal["conversation.item.input_audio_transcription.failed"]
        item_id: str
        content_index: int
        error: ServerEvent.InputAudioTranscriptionError

    class ConversationItemTruncated(TypedDict):
        event_id: str
        type: Literal["conversation.item.truncated"]
        item_id: str
        content_index: int
        audio_end_ms: int

    class ConversationItemDeleted(TypedDict):
        event_id: str
        type: Literal["conversation.item.deleted"]
        item_id: str

    class ResponseCreated(TypedDict):
        event_id: str
        type: Literal["response.created"]
        response: Resource.Response

    class ResponseDone(TypedDict):
        event_id: str
        type: Literal["response.done"]
        response: Resource.Response

    class ResponseOutputItemAdded(TypedDict):
        event_id: str
        type: Literal["response.output_item.added"]
        response_id: str
        output_index: int
        item: Resource.Item

    class ResponseOutputItemDone(TypedDict):
        event_id: str
        type: Literal["response.output.done"]
        response_id: str
        output_index: int
        item: Resource.Item

    class ResponseContentPartAdded(TypedDict):
        event_id: str
        type: Literal["response.content_part.added"]
        item_id: str
        response_id: str
        output_index: int
        content_index: int
        part: ContentPart

    class ResponseContentPartDone(TypedDict):
        event_id: str
        type: Literal["response.content.done"]
        response_id: str
        output_index: int
        content_index: int
        part: ContentPart

    class ResponseTextDeltaAdded(TypedDict):
        event_id: str
        type: Literal["response.text.delta"]
        response_id: str
        output_index: int
        content_index: int
        delta: str

    class ResponseTextDone(TypedDict):
        event_id: str
        type: Literal["response.text.done"]
        response_id: str
        output_index: int
        content_index: int
        text: str

    class ResponseAudioTranscriptDelta(TypedDict):
        event_id: str
        type: Literal["response.audio_transcript.delta"]
        response_id: str
        output_index: int
        content_index: int
        delta: str

    class ResponseAudioTranscriptDone(TypedDict):
        event_id: str
        type: Literal["response.audio_transcript.done"]
        response_id: str
        output_index: int
        content_index: int
        transcript: str

    class ResponseAudioDelta(TypedDict):
        event_id: str
        type: Literal["response.audio.delta"]
        response_id: str
        output_index: int
        content_index: int
        delta: str  # b64

    class ResponseAudioDone(TypedDict):
        event_id: str
        type: Literal["response.audio.done"]
        response_id: str
        output_index: int
        content_index: int

    class ResponseFunctionCallArgumentsDelta(TypedDict):
        event_id: str
        type: Literal["response.function_call_arguments.delta"]
        response_id: str
        output_index: int
        delta: str

    class ResponseFunctionCallArgumentsDone(TypedDict):
        event_id: str
        type: Literal["response.function_call_arguments.done"]
        response_id: str
        name: str
        output_index: int
        call_id: str
        arguments: str

    class RateLimitsData(TypedDict):
        name: Literal["requests", "tokens", "input_tokens", "output_tokens"]
        limit: int
        remaining: int
        reset_seconds: float

    class RateLimitsUpdated:
        event_id: str
        type: Literal["rate_limits.updated"]
        limits: list[ServerEvent.RateLimitsData]


ClientEvents = Union[
    ClientEvent.SessionUpdate,
    ClientEvent.InputAudioBufferAppend,
    ClientEvent.InputAudioBufferCommit,
    ClientEvent.InputAudioBufferClear,
    ClientEvent.ConversationItemCreate,
    ClientEvent.ConversationItemTruncate,
    ClientEvent.ConversationItemDelete,
    ClientEvent.ResponseCreate,
    ClientEvent.ResponseCancel,
]

ServerEvents = Union[
    ServerEvent.Error,
    ServerEvent.SessionCreated,
    ServerEvent.SessionUpdated,
    ServerEvent.ConversationCreated,
    ServerEvent.InputAudioBufferCommitted,
    ServerEvent.InputAudioBufferCleared,
    ServerEvent.InputAudioBufferSpeechStarted,
    ServerEvent.InputAudioBufferSpeechStopped,
    ServerEvent.ConversationItemCreated,
    ServerEvent.ConversationItemInputAudioTranscriptionCompleted,
    ServerEvent.ConversationItemInputAudioTranscriptionFailed,
    ServerEvent.ConversationItemTruncated,
    ServerEvent.ConversationItemDeleted,
    ServerEvent.ResponseCreated,
    ServerEvent.ResponseDone,
    ServerEvent.ResponseOutputItemAdded,
    ServerEvent.ResponseOutputItemDone,
    ServerEvent.ResponseContentPartAdded,
    ServerEvent.ResponseContentPartDone,
    ServerEvent.ResponseTextDeltaAdded,
    ServerEvent.ResponseTextDone,
    ServerEvent.ResponseAudioTranscriptDelta,
    ServerEvent.ResponseAudioTranscriptDone,
    ServerEvent.ResponseAudioDelta,
    ServerEvent.ResponseAudioDone,
    ServerEvent.ResponseFunctionCallArgumentsDelta,
    ServerEvent.ResponseFunctionCallArgumentsDone,
    ServerEvent.RateLimitsUpdated,
]


ClientEventType = Literal[
    "session.update",
    "input_audio_buffer.append",
    "input_audio_buffer.commit",
    "input_audio_buffer.clear",
    "conversation.item.create",
    "conversation.item.truncate",
    "conversation.item.delete",
    "response.create",
    "response.cancel",
]

ServerEventType = Literal[
    "error",
    "session.created",
    "session.updated",
    "conversation.created",
    "input_audio_buffer.committed",
    "input_audio_buffer.cleared",
    "input_audio_buffer.speech_started",
    "input_audio_buffer.speech_stopped",
    "conversation.item.created",
    "conversation.item.input_audio_transcription.completed",
    "conversation.item.input_audio_transcription.failed",
    "conversation.item.truncated",
    "conversation.item.deleted",
    "response.created",
    "response.done",
    "response.output_item.added",
    "response.output_item.done",
    "response.content_part.added",
    "response.content_part.done",
    "response.text.delta",
    "response.text.done",
    "response.audio_transcript.delta",
    "response.audio_transcript.done",
    "response.audio.delta",
    "response.audio.done",
    "response.function_call_arguments.delta",
    "response.function_call_arguments.done",
    "rate_limits.updated",
]



import asyncio
import logging
import uuid
from typing import List, Optional

import websockets
from google import genai
from google.genai import types
from google.genai.live import AsyncSession
from pydantic.v1.main import BaseModel

from ai01.agent._models import AgentsEvents
from ai01.agent.agent import Agent
from ai01.providers.gemini.conversation import Conversation

from ...utils.emitter import EnhancedEventEmitter

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class GeminiConfig(BaseModel):
    function_declaration: Optional[List] = None
    retrieval: Optional[types.Retrieval] = None
    code_execution: Optional[types.ToolCodeExecution] = None
    google_search: Optional[types.GoogleSearch] = None
    google_search_retrieval: Optional[types.GoogleSearchRetrieval] = None

    class Config:
        arbitrary_types_allowed = True


class GeminiOptions(BaseModel):
    """
    realtimeModelOptions is the configuration for the realtimeModel
    """

    gemini_api_key: str
    """
    Gemini API Key is the API Key for the Gemini Provider
    """

    model = "gemini-2.0-flash-exp"
    """
    Model is the Model which is going to be used by the realtimeModel
    """

    system_instruction: Optional[str] = (
        "You are a Helpul Voice Assistant. You can help me with my queries."
    )

    config: GeminiConfig

    """
    Config is the Config which the Model is going to use for the conversation
    """

    class Config:
        arbitrary_types_allowed = True


class GeminiRealtime(EnhancedEventEmitter):
    def __init__(self, agent: Agent, options: GeminiOptions):
        self.agent = agent
        self._options = options

        self.config = types.LiveConnectConfig(
            response_modalities=["AUDIO"],
            system_instruction=types.Content(
                parts=[
                    types.Part(
                        text=options.system_instruction,
                    )
                ]
            ),
            tools=[
                types.Tool(
                    # ignore the below error
                    function_declarations=options.config.function_declaration,
                    google_search=options.config.google_search,
                    google_search_retrieval=options.config.google_search_retrieval,
                    code_execution=options.config.code_execution,
                    retrieval=options.config.retrieval,
                )
            ],
        )

        self.client = genai.Client(
            api_key=self._options.gemini_api_key,
            http_options={"api_version": "v1alpha"},
        )

        self.loop = asyncio.get_event_loop()

        self._logger = logger.getChild(f"realtimeModel-{self._options.model}")

        self.conversation: Conversation = Conversation(id=str(uuid.uuid4()))

        self.session: AsyncSession | None = None

        self.tasks = []

    def __str__(self):
        return f"Gemini realtime: {self._options.model}"

    def __repr__(self):
        return f"Gemini realtime: {self._options.model}"

    async def send_audio(self, audio_bytes: bytes):
        if self.session is None:
            raise Exception("Session is not connected")

        try:
            input = types.LiveClientRealtimeInput(
                media_chunks=[types.Blob(data=audio_bytes, mime_type="audio/pcm")]
            )
            await self.session.send(input=input, end_of_turn=False)
        # {"data": audio_bytes, "mime_type": "audio/pcm"}, end_of_turn=False
        except websockets.exceptions.ConnectionClosed:
            self._logger.warning("WebSocket connection closed while sending audio.")
            self.session = None

    async def handle_response(self):
        try:
            if self.session is None or self.agent.audio_track is None:
                raise Exception("Session or AudioTrack is not connected")
            while True:
                async for response in self.session.receive():
                    if response.data:
                        self.agent.audio_track.enqueue_audio(response.data)
                    elif response.text:
                        print(response.text, end="", flush=True)
                    elif response.tool_call:
                        print("tool call recieved", response.tool_call)
                        self.agent.emit(
                            AgentsEvents.ToolCall, self.session.send, response.tool_call
                        )

        except websockets.exceptions.ConnectionClosedOK:
            self._logger.info("WebSocket connection closed normally.")
        except Exception as e:
            self._logger.error(f"Error in handle_response: {e}")
            raise e

    async def fetch_audio_from_rtc(self):
        while True:
            if not self.conversation.active:
                await asyncio.sleep(0.01)
                continue

            audio_chunk = self.conversation.recv()

            if audio_chunk is None:
                await asyncio.sleep(0.01)
                continue

            await self.send_audio(audio_chunk)

    async def run(self):
        while True:
            try:
                async with self.client.aio.live.connect(
                    model=self._options.model, config=self.config
                ) as session:
                    self.session = session

                    handle_response_task = asyncio.create_task(self.handle_response())
                    fetch_audio_task = asyncio.create_task(self.fetch_audio_from_rtc())

                    self.tasks.extend([handle_response_task, fetch_audio_task])
                    await asyncio.gather(*self.tasks)
            except Exception as e:
                self._logger.error(f"Error in connecting to the Gemini Model: {e}")
                await asyncio.sleep(5)  # Wait before attempting to reconnect

    async def connect(self):
        self.loop.create_task(self.run())



# AgentsEvents is the Enum for the different types of Events emitted by the Agent.
class AgentsEvents(str):
    Connected: str = "Connected"
    Disconnected: str = "Disconnected"
    Speaking: str = "Speaking"
    Listening: str = "Listening"
    Thinking: str = "Thinking"
    ToolCall: str = "ToolCall"



import asyncio
import json
import logging
from typing import Any, Dict, Optional

import websockets

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SocketClient:
    """
    Handles all the websocket related operations
    """

    def __init__(
        self,
        url: str,
        headers: Dict[str, str],
        loop: asyncio.AbstractEventLoop,
        json: bool = True,
    ):
        # URL is the WebSocket server URL.
        self.url = url

        # Headers to be sent with the WebSocket request.
        self.headers = headers

        # Event loop to use for async operations.
        self.loop = loop

        # WebSocket connection instance.
        self.__ws: Optional[websockets.WebSocketClientProtocol] = None

        # Logger for SocketClient.
        self._logger = logger.getChild("socketClient")

        # JSON flag to determine if the messages are JSON or not.
        self.json = json

    @property
    def ws(self) -> websockets.WebSocketClientProtocol:
        # Get the WebSocket connection, raising an error if not connected.
        if self.__ws is None:
            raise Exception("WebSocket is not connected")
        return self.__ws

    @property
    def connected(self) -> bool:
        # Return the connection status of the WebSocket.
        return self.__ws is not None and self.__ws.open

    async def connect(self):
        """
        Connect to the WebSocket server.
        """
        try:
            self._logger.info(f"Attempting to connect to WebSocket at {self.url}")

            self.__ws = await websockets.connect(self.url, extra_headers=self.headers)

            self._logger.info("WebSocket connection established")
        except Exception as e:
            self._logger.error(f"Error connecting to WebSocket: {e}")
            raise

    async def send(self, message: Any):
        """
        Send a message to the WebSocket server.
        """
        try:
            if not self.__ws:
                raise Exception("WebSocket is not connected")

            dump_data = json.dumps(message) if self.json else message

            await self.__ws.send(dump_data)

        except Exception as e:
            self._logger.error(f"Error sending message: {e}")
            raise

    def close(self):
        """
        Close the WebSocket connection.
        """
        if self.__ws:
            self._logger.info("Closing WebSocket connection")
            self.loop.create_task(self.__ws.close())



import json
import logging

from huddle01 import (
    AccessToken,
    AccessTokenData,
    AccessTokenOptions,
    HuddleClient,
    HuddleClientOptions,
    Role,
)
from huddle01.local_peer import ProduceOptions
from pydantic import BaseModel

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger("RTC")


class RTCOptions(BaseModel):
    """
    RTCOptions is the configuration for the Room.
    """

    project_id: str
    """
    Project ID is the Huddle01 Project ID.
    """

    api_key: str
    """
    API Key is the Huddle01 API Key.
    """

    room_id: str
    """
    Room ID is the unique identifier of the Room.
    """

    huddle_client_options: HuddleClientOptions
    """
    Huddle Client Options is the configuration for the Huddle Client.
    """

    metadata: dict[str, str]
    """
    Metadata is the optional metadata for the Room.
    """

    role: str
    """
    Role is the role of the local user in the Room.
    """


class RTC:
    """
    RTC Handles anything related to Real Time Communication, using the dRTC Network,
    Currently it is using Huddle01 as the dRTC Network.
    """

    def __init__(self, options: RTCOptions):
        self._logger = logger.getChild(options.room_id)

        self._options = options

        self._huddle_client = HuddleClient(
            project_id=self._options.project_id,
            options=self._options.huddle_client_options,
        )

    def __str__(self):
        return f"RTC Room: {self._options.room_id}"

    def __repr__(self):
        return f"RTC Room: {self._options.room_id}"

    @property
    def options(self):
        return self._options

    @property
    def huddle_client(self):
        return self._huddle_client

    @property
    def room(self):
        return self._huddle_client.room

    async def produce(self, options: ProduceOptions):
        """
        Produces Media Stream Tracks to the Room.
        """
        local_peer = self.huddle_client.local_peer

        if not local_peer:
            raise ValueError(
                "Local Peer is not created, make sure to connect to the Room before producing."
            )

        await local_peer.produce(options=options)

    async def join(self):
        """
        Joins the dRTC Network and creates a Room for the local user.

        This method establishes a connection to the dRTC Network and creates a Room instance for the local user.
        The returned Room object allows you to listen to events such as:
        - Remote peers joining or leaving the Room.
        - Room-specific events triggered during its lifecycle.

        Example Usage:
            ```python
            room = await rtc.join()


            @room.on(RoomEvents.RoomJoined)
            def on_room_join():
                print("Room successfully joined!")
            ```

        Parameters:
            None (all required options should be configured in `self._options`).

        Returns:
            Room: An instance representing the created Room, enabling event handling and interaction.

        Exceptions:
            Raises an Exception if there is an error while joining the Room.
            Common issues may include:
            - Invalid API key or room ID.
            - Network connectivity problems.
            - Issues during token generation or Room creation.

        Notes:
            - Ensure that `self._options` is correctly configured with the required fields:
                - `room_id`: The unique identifier of the Room.
                - `api_key`: Your Huddle01 API key.
                - `metadata`: Optional metadata for the Room (must be JSON serializable).
                - `role`: The role of the local user in the Room (e.g., "host", "guest").
        """
        self._logger.info("Join Huddle01 dRTC Network")

        accessTokenData = AccessTokenData(
            room_id=self._options.room_id,
            api_key=self._options.api_key,
            options=AccessTokenOptions(
                metadata=json.dumps(self._options.metadata),
            ),
            role=Role(self._options.role),
        )

        accessToken = AccessToken(
            data=accessTokenData,
        )

        try:
            token = await accessToken.to_jwt()

            room = await self.huddle_client.create(self._options.room_id, token)

            return room
        except Exception as e:
            logger.error("Error while joining the Room: ", e)
            raise e



from . import _api as api
from ._api import (
    AudioFormat,
    ClientEvent,
    ClientEventType,
    Modality,
    RealTimeModels,
    ServerEvent,
    ServerEventType,
    ToolChoice,
    Voice,
)
from .realtime_model import RealTimeModel, RealTimeModelOptions

__all__ = [
    "api",
    "RealTimeModel",
    "RealTimeModels",
    "RealTimeModelOptions",
    "ClientEvent",
    "ServerEvent",
    "Voice",
    "Modality",
    "AudioFormat",
    "ToolChoice",
    "ClientEventType",
    "ServerEventType",
]


# Cleanup docs of unexported modules
_module = dir()
NOT_IN_ALL = [m for m in _module if m not in __all__]

__pdoc__ = {}

for n in NOT_IN_ALL:
    __pdoc__[n] = False
